#!/bin/bash
#
#  Script to generate configuration files for CentOS7 configuration
#
#
#  When installing files, use
#     cp --backup=numbered --verbose
VERBOSE=1
COLOR=1
DISKLESS_CFG=diskless.cfg
DO_INSTALL=1
SKIP_GENERATE=0
FORCE_INSTALL=0
MKIMAGE=bin/mkenvimage
OUTPUT_PATH=output
NFS_SERVER_HOSTNAME=$(hostname -s)

#
# Install command and its default arguments
#   add -v for verbose
#
#INSTALL=''
INSTALL='install_func'

#
# NFS server
#   /etc/exports
#

CONFIG_VERBOSE() {
    if [ "$VERBOSE" -eq "1" ]; then
	if [ "$COLOR" -eq "1" ]; then
	    echo -en '\E[;33m'
	fi
	echo $@
	if [ "$COLOR" -eq "1" ]; then
	    echo -en '\E[;36m'
	fi
    fi
}

CONFIG_INFO() {
    if [ "$COLOR" -eq "1" ]; then
	echo -en '\E[;33m'
    fi
    echo $@
    if [ "$COLOR" -eq "1" ]; then
	echo -en '\E[;36m'
    fi
}

CONFIG_ERROR() {
    if [ "$COLOR" -eq "1" ]; then
	echo -en '\E[;31m'
    fi
    echo $@ >&2
    if [ "$COLOR" -eq "1" ]; then
	echo -en '\E[;36m'
    fi
}

install_func() {
    echo "INSTALL: $@"
    if [ $1 = "-d" ]; then
	install -b -p -m 755 $@
    else
	install -b -p -m 644 $@
    fi
}

check_output_path() {
    if [ ! -d ${OUTPUT_PATH} ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Generating OUTPUT_PATH=${OUTPUT_PATH}"
	CONFIG_VERBOSE
	mkdir -p ${OUTPUT_PATH}
    fi

}

generate_nfs_exports() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Generating NFS exports"
    CONFIG_VERBOSE
    OUTPUT_FILE=${OUTPUT_PATH}/${NFS_EXPORTS}
    OUT=

    check_output_path

    # root
    OUT[0]="${NFS_ROOT_PATH}   "
    for element in $(seq 0 $((${#CLIENT_HOSTNAME[@]} - 1)))
    do
	OUT[$element+1]="${CLIENT_HOSTNAME[$element]}(rw,no_root_squash,async)"
    done
    echo ${OUT[@]} > $OUTPUT_FILE

    # snapshot
    OUT[0]="${NFS_SNAPSHOT_PATH}   "
    for element in $(seq 0 $((${#CLIENT_HOSTNAME[@]} - 1)))
    do
	OUT[$element+1]=${CLIENT_HOSTNAME[$element]}"(rw,no_root_squash,async)"
    done
    echo ${OUT[@]} >> $OUTPUT_FILE

    # home
    OUT[0]="${NFS_HOME_PATH}   "
    for element in $(seq 0 $((${#CLIENT_HOSTNAME[@]} - 1)))
    do
	OUT[$element+1]=${CLIENT_HOSTNAME[$element]}"(rw,no_root_squash,async)"
    done
    echo ${OUT[@]} >> $OUTPUT_FILE

    CONFIG_INFO "  ... Generated: "${OUTPUT_FILE}
}

install_nfs_exports() {
    generate_nfs_exports

    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Installing NFS exports:"
	CONFIG_VERBOSE

	# Warning if file already exists
	if [ -e /etc/exports.d/${NFS_EXPORTS} ]; then
	    CONFIG_INFO "WARNING: /etc/exports.d/${NFS_EXPORTS} exists. Backup creation enabled."
	fi

	${INSTALL} $OUTPUT_PATH/${NFS_EXPORTS} /etc/exports.d/${NFS_EXPORTS}
    fi
}


#
# Client root
#   /etc/yp.conf
#

generate_yp_conf() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Generating yp.conf"
    CONFIG_VERBOSE

    check_output_path

    # Just copy the file from template to output dir
    OUTPUT_FILE=${OUTPUT_PATH}/yp.conf
    cp $YP_CONF_TEMPLATE $OUTPUT_FILE

    CONFIG_INFO "  ... Generated: "${OUTPUT_FILE}
}

install_yp_conf() {
    generate_yp_conf

    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Installing yp.conf"
	CONFIG_VERBOSE

	${INSTALL} $OUTPUT_PATH/yp.conf $NFS_ROOT_PATH/etc/yp.conf
    fi
}

#
# Client root
#   /etc/auto.master.d/home.autofs
#

generate_autofs_home() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Generating home.autofs"
    CONFIG_VERBOSE

    check_output_path

    OUTPUT_FILE=${OUTPUT_PATH}/auto.home
    TEMPLATE=${AUTOFS_HOME_TEMPLATE}
    sed -e "s|NFS_SERVER_HOSTNAME|"${NFS_SERVER_HOSTNAME}"|g" \
	-e "s|NFS_HOME_PATH|"${NFS_HOME_PATH}"|g" < ${TEMPLATE} > $OUTPUT_FILE

    CONFIG_INFO "  ... Generated: "${OUTPUT_FILE}
}

install_autofs_home() {
    generate_autofs_home

    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Installing auto.home"
	CONFIG_VERBOSE

	${INSTALL} ${TEMPLATE_PATH}/home.autofs ${NFS_ROOT_PATH}/etc/auto.master.d/home.autofs
	${INSTALL} ${OUTPUT_PATH}/auto.home ${NFS_ROOT_PATH}/etc/auto.master.d/auto.home
    fi
}

#
# Client root
#   /etc/sysconfig/readonly-root
#

generate_readonly_root() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Generating readonly-root"
    CONFIG_VERBOSE

    check_output_path

    # /etc/sysconfig/readonly-root override
    CLIENTSTATE=${NFS_SERVER_HOSTNAME}:${NFS_SNAPSHOT_PATH}

    OUTPUT_FILE=${OUTPUT_PATH}/readonly-root
    TEMPLATE=${READONLY_ROOT_TEMPLATE}
    sed "s|CLIENTSTATE=.*|CLIENTSTATE="$CLIENTSTATE"|g" < ${TEMPLATE} > $OUTPUT_FILE

    CONFIG_INFO "  ... Generated: "${OUTPUT_FILE}
}

install_readonly_root() {
    generate_readonly_root

    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Installing readonly-root"
	CONFIG_VERBOSE

	${INSTALL} ${OUTPUT_PATH}/readonly-root ${NFS_ROOT_PATH}/etc/sysconfig/readonly-root
    fi

}

#
# TFTP PXE boot files
#   $tftpboot-dir/linux-install/pxelinux.cfg/$ADDRFILE
#
#

generate_pxelinux_boot() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Generating pxelinux config files"
    CONFIG_VERBOSE

    # Bail if the VTP MAC address is defined.  Because it doesnt need pxelinux
    if [ ! -z "$VTP_MAC_ADDR" ] ; then
	CONFIG_VERBOSE "Skipping, because VTP_MAC_ADDR is defined"
	return;
    fi

    check_output_path

    # Check for gethostip executable
    CHECKPLS=$(which gethostip 2> /dev/null)
    if [ "$?" -eq 1 ]; then
       CONFIG_ERROR "ERROR:  Missing executable 'gethostip'"
       CONFIG_ERROR "        Install syslinux package to correct"
       return
    fi

    OUTPUT_FILE=${OUTPUT_PATH}/${TFTP_PXELINUX_CFG_DEFAULT}
    TEMPLATE=${PXECONFIG_ADDR_TEMPLATE}
    TFTP_KERNEL_DIR=$(basename ${TFTP_KERNEL_PATH})
    NFS_SERVER_ADDR=$(gethostip -d ${NFS_SERVER_HOSTNAME})

    sed -e "s|TFTP_KERNEL_DIR|"${TFTP_KERNEL_DIR}"|g" \
	-e "s|VMLINUZ|"${VMLINUZ}"|g" \
	-e "s|INITRD|"${INITRD}"|g" \
	-e "s|NFS_SERVER_ADDR|"${NFS_SERVER_ADDR}"|g" \
	-e "s|NFS_ROOT_PATH|"${NFS_ROOT_PATH}"|g" < ${TEMPLATE} > $OUTPUT_FILE

    # Loop over client hostnames to generate sym-links
    for element in $(seq 0 $((${#CLIENT_HOSTNAME[@]} - 1)))
    do
	ADDR_FILENAME=$(gethostip -x ${CLIENT_HOSTNAME[$element]})
	if [ "$?" -eq 1 ]; then
	    CONFIG_ERROR "ERROR:  Invalid hostname"
	    CONFIG_ERROR "        ${CLIENT_HOSTNAME[$element]}"
	    return
	fi

	OUTPUT_FILE=${OUTPUT_PATH}/${ADDR_FILENAME}
	ln -sf ${TFTP_PXELINUX_CFG_DEFAULT} ${OUTPUT_FILE}

	CONFIG_INFO "  ... Linked: ${OUTPUT_FILE} to ${TFTP_PXELINUX_CFG_DEFAULT}"
    done
}

install_pxelinux_boot() {
    generate_pxelinux_boot

    # Bail if the VTP MAC address is defined.  Because it doesnt need pxelinux
    if [ ! -z "$VTP_MAC_ADDR" ] ; then
	CONFIG_VERBOSE "Skipping, because VTP_MAC_ADDR is defined"
	return;
    fi

    if [ "$DO_INSTALL" == "1" ]; then

	CONFIG_VERBOSE
	CONFIG_VERBOSE "Installing pxelinux config files"
	CONFIG_VERBOSE

	# Create the output path if it does not exist
	if [ ! -d ${TFTP_PXELINUX_CFG_PATH} ]; then
	    mkdir -v -p ${TFTP_PXELINUX_CFG_PATH}
	fi

	# Copy over the default file
	${INSTALL} ${OUTPUT_PATH}/${TFTP_PXELINUX_CFG_DEFAULT} \
		   ${TFTP_PXELINUX_CFG_PATH}/

	# Loop over client hostnames to generate sym-links
	for element in $(seq 0 $((${#CLIENT_HOSTNAME[@]} - 1)))
	do
	    ADDR_FILENAME=$(gethostip -x ${CLIENT_HOSTNAME[$element]})
	    if [ "$?" -eq 1 ]
	    then
		CONFIG_ERROR "ERROR:  Invalid hostname"
		CONFIG_ERROR "        ${CLIENT_HOSTNAME[$element]}"
		return
	    fi

	    OUTPUT_FILE=${TFTP_PXELINUX_CFG_PATH}/${ADDR_FILENAME}
	    ln -sf ${TFTP_PXELINUX_CFG_DEFAULT} ${OUTPUT_FILE}

	    CONFIG_INFO "  ... Linked: ${OUTPUT_FILE} to ${TFTP_PXELINUX_CFG_DEFAULT}"
	done
    fi
}

#
# u-boot environment file
#   must be copied to VTPs microSD
#
#

generate_uboot_env() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Generating U-Boot Environment"
    CONFIG_VERBOSE

    check_output_path

    # Check for mkenvimage executable
    if [ ! -e $MKIMAGE ]; then
       CONFIG_ERROR "ERROR:  Missing executable bin/mkenvimage"
       CONFIG_ERROR ""
       return
    fi

    # Check for gethostip executable
    CHECKPLS=$(which gethostip 2> /dev/null)
    if [ "$?" -eq 1 ]; then
       CONFIG_ERROR "ERROR:  Missing executable 'gethostip'"
       CONFIG_ERROR "        Install syslinux package to correct"
       return
    fi

    NFS_SERVER_ADDR=$(gethostip -d ${NFS_SERVER_HOSTNAME})
    if [ "$?" -eq 1 ]
    then
	CONFIG_ERROR "ERROR:  Invalid hostname"
	CONFIG_ERROR "        ${NFS_SERVER_NAME}"
	return
    fi


    # Generate env.txt from template
    OUTPUT_FILE=${OUTPUT_PATH}/env.txt
    TEMPLATE=${UBOOT_ENV_TEMPLATE}

    sed -e "s|NFS_SERVER_ADDR|"${NFS_SERVER_ADDR}"|g" \
	-e "s|VTP_MAC_ADDR|"${VTP_MAC_ADDR}"|g" \
	-e "s|NFS_ROOT_PATH|"${NFS_ROOT_PATH}"|g" < ${TEMPLATE} > $OUTPUT_FILE

    # Convert env.txt to env.bin
    $MKIMAGE -s 0x20000 -o ${OUTPUT_PATH}/env.bin ${OUTPUT_PATH}/env.txt
    local rval=$?
    if [ "$rval" != "0" ]; then
	CONFIG_ERROR "ERROR: mkenvimage returned an error ($rval)"
	CONFIG_ERROR ""
	return
    fi

    CONFIG_INFO " ${OUTPUT_PATH}/env.bin now ready for VTP microSD"

}

#
# Client snapshot
#
#
install_client_snapshot() {
    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Installing client snapshot"
	CONFIG_VERBOSE

	# Loop over client hostnames to generate client snapshot directories
	for element in $(seq 0 $((${#CLIENT_HOSTNAME[@]} - 1)))
	do
	    ${INSTALL} -d ${NFS_SNAPSHOT_PATH}/${CLIENT_HOSTNAME[$element]}
	done
    fi
}


#
# Remove the additions that are specific to the JLab CUE
# This removes the mounts of /u (site and apps)
# and disables the ypbind service (handles the CUE login credentials)
#

remove_jlab_additions() {
    CONFIG_VERBOSE
    CONFIG_VERBOSE "Removing JLAB additions"
    CONFIG_VERBOSE

    if [ ! -d "$NFS_ROOT_PATH" ]; then
	CONFIG_ERROR "ERROR: NFS_ROOT_PATH does not exist"
	CONFIG_ERROR "    extract TGZ_ROOT first (diskless_gen.sh -r )"
    fi


    # Disable the ypbind.service (remove the softlink)
    RMFILE=${NFS_ROOT_PATH}/etc/systemd/system/multi-user.target.wants/ypbind.service
    if [ -e $RMFILE ]; then
	rm -v $RMFILE
    fi

    # Remove the yp.conf
    RMFILE=${NFS_ROOT_PATH}/etc/yp.conf
    if [ -e $RMFILE ]; then
	rm -v $RMFILE
    fi

}

#
# Untar the ARCH tftpboot files into the specified area
#
#

install_tgz_tftp() {
    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Extracting TGZ_TFTP=$TGZ_TFTP to "
	CONFIG_VERBOSE "TFTP_KERNEL_PATH=$TFTP_KERNEL_PATH"
	CONFIG_VERBOSE

	if [ ! -e ${TGZ_TFTP} ]; then
	    CONFIG_ERROR
	    CONFIG_ERROR "ERROR: TGZ_TFTP not found"
	    CONFIG_ERROR "TGZ_TFTP=$TGZ_TFTP"
	    CONFIG_ERROR
	    return
	fi

	if [ ! -d "$TFTP_KERNEL_PATH" ]; then
	    install -d -b ${TFTP_KERNEL_PATH}
	else
	    if [ "$FORCE_INSTALL" == "1" ]; then
		rm -r ${TFTP_KERNEL_PATH}
		install -d -b ${TFTP_KERNEL_PATH}
	    else
		CONFIG_ERROR
		CONFIG_ERROR "ERROR: TFTP_KERNEL_PATH already exists. "
		CONFIG_ERROR "       Remove before installing tftp files"
		CONFIG_ERROR "TFTP_KERNEL_PATH=$TFTP_KERNEL_PATH"
		CONFIG_ERROR
		return
	    fi
	fi

	tar xzvf ${TGZ_TFTP} -C ${TFTP_KERNEL_PATH}
    fi
}

install_tgz_root() {
    if [ "$DO_INSTALL" == "1" ]; then
	CONFIG_VERBOSE
	CONFIG_VERBOSE "Extracting TGZ_ROOT=$TGZ_ROOT to "
	CONFIG_VERBOSE "NFS_ROOT_PATH=$NFS_ROOT_PATH"
	CONFIG_VERBOSE

	if [ ! -d "$NFS_ROOT_PATH" ]; then
	    mkdir -p ${NFS_ROOT_PATH}
	else
	    if [ "$FORCE_INSTALL" == "1" ]; then
		/bin/rm -rf ${NFS_ROOT_PATH}
		install -d -b ${NFS_ROOT_PATH}
	    else
		CONFIG_ERROR
		CONFIG_ERROR "ERROR: NFS_ROOT_PATH already exists. "
		CONFIG_ERROR "       Remove before installing tftp files"
		CONFIG_ERROR "NFS_ROOT_PATH=$NFS_ROOT_PATH"
		CONFIG_ERROR
		return
	    fi
	fi

	tar xzf ${TGZ_ROOT} -C ${NFS_ROOT_PATH}

	# Make the base snapshot and home directories now, if we need to
	if [ ! -d ${NFS_SNAPSHOT_PATH} ]; then
	    install -d -b ${NFS_SNAPSHOT_PATH}
	fi

    	if [ ! -d ${NFS_HOME_PATH} ]; then
	    install -d -b ${NFS_HOME_PATH}
	fi
fi
}

usage()
{
    cat << EOF >&2

Usage: diskless_gen.sh [OPTION...] [CLIENT]...
'diskless_gen.sh' generates and installs configuration files for diskless OS

Examples:
   diskless_gen.sh         # generate and install all diskless config files
   diskless_gen.sh daroc1  # generate and install client config files
                           #  for client with hostname 'daroc1'
   diskless_gen.sh -g      # generate config files without installation
                           #  output into $OUTPUT_PATH

Configuration file definitions (default file: diskless.cfg)
   OUTPUT_PATH                # Where to output generated config files

   CLIENT_HOSTNAME[0]         # Bash array of client hostnames
   CLIENT_HOSTNAME[1]
   ...
   CLIENT_HOSTNAME[n]

   NFS_SERVER_HOSTNAME        # Hostname of NFS server
   NFS_PATH                   # Local base PATH for NFS exports (root, home, and snapshot)
   NFS_ROOT_PATH              # root PATH
   NFS_SNAPSHOT_PATH	      # snapshot PATH
   NFS_HOME_PATH              # home PATH

   TFTP_PXELINUX_CFG_PATH     # Local base PATH for tftp pxelinux.cfg directory
   TFTP_PXELINUX_CFG_DEFAULT  # Filename of default file in pxelinux.cfg
   TFTP_KERNEL_PATH            # Local PATH of boot kernel files

   TEMPLATE_PATH              # PATH For diskless_gen.sh template files
   YP_CONF_TEMPLATE           # yp.conf template
   READONLY_ROOT_TEMPLATE     # readonly-root template
   PXECONFIG_ADDR_TEMPLATE    # pxelinux.cfg address file template
   AUTOFS_HOME_TEMPLATE       # home.autofs template

   CLIENTSTATE                # Mount point for snapshot directory (NFS or local)

   TGZ_TFTP                   # kernel boot file archive for tftp
   TGZ_ROOT                   # OS root directory archive

Output/install Options:

   -g                         generate config files without installation
   -o                         install the config files in OUTPUT_PATH

Config file Option:

   -f FILE                    Use FILE for configuration file definitions
      			      (default diskless.cfg)

Host Specific Options:

   -p                         pxelinux.cfg setup only
   -n			      NFS setup only
   -b                         extract TGZ_TFTP into TFTP_KERNEL_PATH
   -r			      extract TGZ_ROOT into NFS_ROOT_PATH
   -d			      when extracting TGZ, delete older TFTP_KERNEL_PATH
   			       or NFS_ROOT_PATH already exists.

Client OS Options
   -u                         generated (and install) yp.conf, home.autofs, readonly-root
   			      for Client's OS
   -j			      Remove the jlab network specific additions (yp.conf, jlab.autofs)
   -e			      generate env.bin for VTP U-Boot environment

Client Specific Options:

   -c                         process only the client config files
   			      (pxelinux.cfg address files, snapshot)
   -s			      generate only the snapshot directories

Help Options:

   -v                         verbose output
   -l			      verbose and colored output
   -h			      give this help list

EOF

}


# Handle the setup arguments first
while getopts ":gof:vldh" flag
do
    case $flag in
	g)
	    DO_INSTALL=0;;
	o)
	    SKIP_GENERATE=1;;
	f)
	    DISKLESS_CFG=$OPTARG;;
	v)
	    VERBOSE=1;;
	l)
	    VERBOSE=1
	    COLOR=1;;

	d)
	    FORCE_INSTALL=1;;

	h)
	    usage
	    exit;;
    esac
done
OPTIND=1

# Load up variables set in configuration file

if [ -f ${DISKLESS_CFG} ]; then
	. ${DISKLESS_CFG}
else
    echo
    CONFIG_ERROR "ERROR.  Configuration file ${DISKLESS_CFG} is missing"
    usage
    exit
fi


# Process rest of the commandline options
while getopts "gof:Epnbrujecsvldh" flag
do
    case $flag in
	g | o | f | v | l | d) ;; # already handled above

	# Everything
	E)
	    install_tgz_tftp
	    install_tgz_root
	    install_pxelinux_boot
	    install_nfs_exports
	    install_yp_conf
	    install_autofs_home
	    install_readonly_root
	    install_client_snapshot
	    exit;;

	# Host Specific Options:
	p)      # pxelinux.cfg/ setup only
	    install_pxelinux_boot
	    exit;;
	n)      # NFS setup only
	    install_nfs_exports
	    exit;;
	b)      # extract and install TGZ_TFTP only
	    install_tgz_tftp
	    exit;;
	r)      # extract and install TGZ_ROOT only
	    install_tgz_root
	    exit;;

	# Client OS Options
	u)      # Generate and install  yp.conf, home.autofs, readonly-root
	    install_yp_conf
	    install_autofs_home
	    install_readonly_root
	    exit;;
	j)      # Remove jlab additions (ypbind disabled, no jlab.autofs)
	    remove_jlab_additions
	    exit;;
	e)      # Generate VTP env.bin
	    generate_uboot_env
	    exit;;

	# Client Specific Options:
	c)
	    install_pxelinux_boot
	    install_client_snapshot
	    exit;;

	s)
	    install_client_snapshot
	    exit;;

	# Help Options:
	h | ?)
	    usage
	    exit;;
    esac
done
